from __future__ import annotations

from datetime import datetime, timezone
import json
from typing import Any, Iterator

from .csv_utils import render_csv_rows
from ..domain.dedupe import DedupeResult

PLAN_VERSION = 1


def _path_map(values: dict) -> dict[str, str]:
    return {k.as_posix(): v.as_posix() for k, v in values.items()}


def _count_map(values: dict) -> dict[str, int]:
    return {k.as_posix(): v for k, v in values.items()}


def dedupe_result_to_dict(
    result: DedupeResult,
    *,
    roots: list | None = None,
    exts: set | None = None,
) -> dict[str, Any]:
    payload: dict[str, Any] = {
        "version": PLAN_VERSION,
        "generated_at": datetime.now(timezone.utc).isoformat(),
        "redundant": sorted([p.as_posix() for p in result.redundant]),
        "duplicate_of": _path_map(result.duplicate_of),
        "contained_in": _path_map(result.contained_in),
        "unsafe": sorted([p.as_posix() for p in result.unsafe]),
        "sizes": _count_map(result.sizes),
        "unique_count": _count_map(result.unique_count),
        "post_contained": [
            {"subset": rel.subset.as_posix(), "superset": rel.superset.as_posix()}
            for rel in result.post_contained
        ],
    }
    if roots is not None:
        payload["roots"] = [p.as_posix() for p in roots]
    if exts is not None:
        payload["exts"] = sorted(exts)
    return payload


def render_dedupe_csv(result: DedupeResult) -> str:
    rows: list[list[object]] = [["release", "action", "reason", "reference", "tracks", "unique_tracks"]]
    for rel in sorted(result.redundant, key=lambda p: p.as_posix()):
        reason = "duplicate" if rel in result.duplicate_of else "contained"
        reference = result.duplicate_of.get(rel) or result.contained_in.get(rel)
        reference_str = reference.as_posix() if reference is not None else ""
        tracks = result.sizes.get(rel)
        unique = result.unique_count.get(rel)
        rows.append(
            [
                rel.as_posix(),
                "remove",
                reason,
                reference_str,
                tracks if tracks is not None else "",
                unique if unique is not None else "",
            ]
        )
    return render_csv_rows(rows)


def iter_dedupe_jsonl(result: DedupeResult) -> Iterator[str]:
    for rel in sorted(result.redundant, key=lambda p: p.as_posix()):
        reference = result.duplicate_of.get(rel) or result.contained_in.get(rel)
        payload = {
            "release": rel.as_posix(),
            "action": "remove",
            "reason": "duplicate" if rel in result.duplicate_of else "contained",
            "reference": reference.as_posix() if reference is not None else None,
            "tracks": result.sizes.get(rel),
            "unique_tracks": result.unique_count.get(rel),
        }
        yield json.dumps(payload, ensure_ascii=False)



from __future__ import annotations

from pathlib import Path

from ..app.release import ReleaseBuildResult


def _format_rel_path(root: Path, path: Path) -> str:
    try:
        return path.relative_to(root).as_posix()
    except ValueError:
        return path.as_posix()


def render_missing_assets_report(
    result: ReleaseBuildResult,
    root: Path,
    *,
    dr_dir: Path | None,
) -> str:
    lines: list[str] = []
    lines.append(f"Missing assets report for: {root.name}")
    lines.append(f"Root: {root}")
    if dr_dir is None:
        lines.append("DR check: disabled (--dr-dir not set)")
    else:
        lines.append(f"DR directory: {dr_dir}")

    lines.append("")
    if result.missing_covers:
        lines.append(f"Missing cover.jpg: {len(result.missing_covers)}")
        for rel in sorted(result.missing_covers, key=lambda p: p.as_posix()):
            lines.append(f"- {_format_rel_path(root, rel)}")
    else:
        lines.append("Missing cover.jpg: 0")

    lines.append("")
    if result.failed_cover_uploads:
        lines.append(f"Failed cover uploads: {len(result.failed_cover_uploads)}")
        for rel in sorted(result.failed_cover_uploads, key=lambda p: p.as_posix()):
            lines.append(f"- {_format_rel_path(root, rel)}")
    else:
        lines.append("Failed cover uploads: 0")

    lines.append("")
    lines.append(f"Cover uploads skipped: {'yes' if result.cover_uploads_skipped else 'no'}")

    if not result.dr_checked:
        lines.append("Missing DR reports: skipped")
    elif result.missing_drs:
        lines.append(f"Missing DR reports: {len(result.missing_drs)}")
        for rel in sorted(result.missing_drs, key=lambda p: p.as_posix()):
            lines.append(f"- {_format_rel_path(root, rel)}")
    else:
        lines.append("Missing DR reports: 0")

    return "\n".join(lines)



from __future__ import annotations

import json
from typing import Any
from pathlib import Path

from .csv_utils import render_csv_rows
from ..domain.grouping import ReleaseGroup
from ..domain.models import Release, StatsSummary
from ..domain.utils import (
    bit_label,
    format_hhmmss,
    group_key,
    release_word,
    sr_label,
    track_word,
)


def _format_release_path(root: Path, release_path: Path) -> str:
    return release_path.relative_to(root).as_posix()


def _format_track_duration(duration: float | None) -> str:
    return format_hhmmss(duration) if duration is not None else "unknown"


def render_stats_text(
    groups: list[ReleaseGroup],
    summary: StatsSummary,
    root: Path,
    *,
    include_tracks: bool = False,
) -> str:
    lines: list[str] = []

    for i, group in enumerate(groups):
        if i:
            lines.append("")
        lines.append(f"{group.name}:")
        for rel in group.releases:
            pretty = _format_release_path(root, rel.path)
            lines.append(
                f"  {pretty} - {format_hhmmss(rel.duration_seconds)} "
                f"({rel.track_count} {track_word(rel.track_count)}, "
                f"{bit_label(rel.bit_depths)}, {sr_label(rel.sample_rates)})"
            )
            if rel.unreadable_track_count:
                lines.append(f"    Warning: unreadable tracks skipped: {rel.unreadable_track_count}")
            if include_tracks:
                for track in rel.tracks:
                    dur = _format_track_duration(track.duration_seconds)
                    sr = f"{track.sample_rate} Hz" if track.sample_rate is not None else "unknown"
                    bit = f"{track.bit_depth}-bit" if track.bit_depth is not None else "unknown"
                    lines.append(f"    {track.path.name} - {dur} ({sr}, {bit})")

    total_releases = sum(len(group.releases) for group in groups)
    lines.append(
        f"\nTotal: {format_hhmmss(summary.total_seconds)} "
        f"({summary.total_tracks} {track_word(summary.total_tracks)}, "
        f"{total_releases} {release_word(total_releases)})"
    )

    return "\n".join(lines)


def render_stats_json(
    groups: list[ReleaseGroup],
    summary: StatsSummary,
    root: Path,
    *,
    include_tracks: bool = False,
) -> str:
    groups_list: list[dict[str, Any]] = []
    data: dict[str, Any] = {
        "groups": groups_list,
        "summary": {
            "total_seconds": summary.total_seconds,
            "total_duration": format_hhmmss(summary.total_seconds),
            "total_tracks": summary.total_tracks,
            "total_releases": sum(len(group.releases) for group in groups),
            "sample_rates": sorted(summary.total_sr),
            "bit_depths": sorted(summary.total_bit),
            "exts": sorted(summary.total_exts),
            "scanned_audio_files": summary.scanned_audio_files,
            "unreadable_audio_files": summary.unreadable_audio_files,
        },
    }

    for group in groups:
        releases_list: list[dict[str, Any]] = []
        group_item: dict[str, Any] = {"name": group.name, "releases": releases_list}
        for rel in group.releases:
            rel_path = rel.path.relative_to(root).as_posix()
            releases_list.append(
                {
                    "rel_path": rel_path,
                    "display_path": _format_release_path(root, rel.path),
                    "duration_seconds": rel.duration_seconds,
                    "duration": format_hhmmss(rel.duration_seconds),
                    "track_count": rel.track_count,
                    "unreadable_track_count": rel.unreadable_track_count,
                    "sample_rates": sorted(rel.sample_rates),
                    "bit_depths": sorted(rel.bit_depths),
                    "exts": sorted(rel.exts),
                }
            )
            if include_tracks:
                track_items: list[dict[str, Any]] = []
                for track in rel.tracks:
                    duration = track.duration_seconds
                    track_items.append(
                        {
                            "rel_path": track.path.relative_to(root).as_posix(),
                            "file_name": track.path.name,
                            "duration_seconds": duration,
                            "duration": _format_track_duration(duration) if duration is not None else None,
                            "sample_rate": track.sample_rate,
                            "bit_depth": track.bit_depth,
                            "ext": track.ext,
                        }
                    )
                releases_list[-1]["tracks"] = track_items
        groups_list.append(group_item)

    return json.dumps(data, ensure_ascii=False)


def render_stats_csv(releases: list[Release], root: Path) -> str:
    rows: list[list[object]] = [
        [
            "group",
            "rel_path",
            "display_path",
            "duration_seconds",
            "duration",
            "track_count",
            "unreadable_track_count",
            "sample_rates",
            "bit_depths",
            "exts",
        ]
    ]
    for rel in releases:
        rel_path = rel.path.relative_to(root).as_posix()
        group = group_key(rel.path.relative_to(root))
        sample_rates = ";".join(str(v) for v in sorted(rel.sample_rates))
        bit_depths = ";".join(str(v) for v in sorted(rel.bit_depths))
        exts = ";".join(sorted(rel.exts))
        rows.append(
            [
                group,
                rel_path,
                _format_release_path(root, rel.path),
                rel.duration_seconds,
                format_hhmmss(rel.duration_seconds),
                rel.track_count,
                rel.unreadable_track_count,
                sample_rates,
                bit_depths,
                exts,
            ]
        )
    return render_csv_rows(rows)


def render_stats_csv_tracks(releases: list[Release], root: Path) -> str:
    rows: list[list[object]] = [
        [
            "group",
            "rel_path",
            "display_path",
            "track_rel_path",
            "track_name",
            "track_ext",
            "duration_seconds",
            "duration",
            "sample_rate",
            "bit_depth",
        ]
    ]
    for rel in releases:
        rel_path = rel.path.relative_to(root).as_posix()
        group = group_key(rel.path.relative_to(root))
        display = _format_release_path(root, rel.path)
        for track in rel.tracks:
            track_rel = track.path.relative_to(root).as_posix()
            duration = track.duration_seconds
            duration_text = _format_track_duration(duration) if duration is not None else ""
            rows.append(
                [
                    group,
                    rel_path,
                    display,
                    track_rel,
                    track.path.name,
                    track.ext,
                    "" if duration is None else duration,
                    duration_text,
                    "" if track.sample_rate is None else track.sample_rate,
                    "" if track.bit_depth is None else track.bit_depth,
                ]
            )
    return render_csv_rows(rows)



from __future__ import annotations

from contextlib import contextmanager
import importlib
import sys
from typing import Any, Callable, Iterator, TypeVar

from ..app.progress import NullProgress, ProgressCallback


class _RichProgress(ProgressCallback):
    def __init__(self, progress: Any, task_id: Any) -> None:
        self._progress = progress
        self._task_id = task_id
        self._total: int | None = None
        self._finished = False

    def start(self, total: int) -> None:
        self._total = total
        self._finished = False
        self._progress.update(self._task_id, total=total, completed=0)

    def advance(self, step: int = 1) -> None:
        self._progress.advance(self._task_id, step)

    def set_description(self, description: str) -> None:
        self._progress.update(self._task_id, description=description)

    def finish(self) -> None:
        if self._finished:
            return
        self._finished = True
        if self._total is not None:
            self._progress.update(self._task_id, completed=self._total)


@contextmanager
def progress_bar(description: str) -> Iterator[ProgressCallback]:
    try:
        rich_progress = importlib.import_module("rich.progress")
    except Exception:
        yield NullProgress()
        return

    progress = rich_progress.Progress(
        rich_progress.SpinnerColumn(),
        rich_progress.TextColumn("{task.description}"),
        rich_progress.BarColumn(),
        rich_progress.TextColumn("{task.completed}/{task.total}"),
        rich_progress.TimeElapsedColumn(),
        transient=True,
    )
    task_id = progress.add_task(description, total=1)
    with progress:
        tracker = _RichProgress(progress, task_id)
        try:
            yield tracker
        finally:
            tracker.finish()


def progress_enabled(no_progress: bool) -> bool:
    return (not no_progress) and sys.stdout.isatty()


T = TypeVar("T")


def run_with_progress(
    no_progress: bool,
    quiet: bool,
    description: str,
    fn: Callable[[ProgressCallback | None], T],
) -> T:
    if progress_enabled(no_progress) and not quiet:
        with progress_bar(description) as progress:
            return fn(progress)
    return fn(None)



from __future__ import annotations

import argparse
from pathlib import Path

from ..args import add_common_audio_args, add_no_progress_arg, normalize_exts
from ..common import ensure_outside_roots, prepare_audio_root, write_output_text
from ..progress import run_with_progress
from ...app.stats import collect_stats, collect_synthetic_stats
from ...domain.grouping import group_releases
from ...formatting.stats import (
    render_stats_csv,
    render_stats_csv_tracks,
    render_stats_json,
    render_stats_text,
)
from ...infra.ffprobe import FfprobeClient


def add_parser(subparsers) -> argparse.ArgumentParser:
    """Register the stats subcommand parser."""
    parser = subparsers.add_parser("stats", help="Print grouped stats.")
    add_common_audio_args(parser, include_root=True)
    add_no_progress_arg(parser)
    parser.add_argument("--json", action="store_true", help="Output JSON stats.")
    parser.add_argument("--csv", action="store_true", help="Output CSV stats.")
    parser.add_argument(
        "--per-track",
        action="store_true",
        help="Include per-track details in the output.",
    )
    parser.add_argument(
        "--output",
        default=None,
        help="Write output to a file instead of stdout.",
    )
    parser.add_argument(
        "--synthetic",
        action="store_true",
        help="Generate fake data (no ffprobe/files needed) to test output formatting.",
    )
    return parser


def run(args: argparse.Namespace) -> int:
    """Execute the stats command."""
    if args.json and args.csv:
        print("Error: --json and --csv cannot be used together.")
        return 2

    root, err = prepare_audio_root(args.root, skip_checks=args.synthetic)
    if err is not None:
        return err
    out_path: Path | None = None
    if args.output is not None:
        out_path = Path(args.output).expanduser().resolve()
        if not ensure_outside_roots(out_path, [root], "output file"):
            return 2

    exts = normalize_exts(args.ext)
    if args.synthetic:
        releases, summary = collect_synthetic_stats(root)
    else:
        ffprobe = FfprobeClient()
        releases, summary = run_with_progress(
            args.no_progress,
            args.json or args.csv,
            "Reading audio metadata",
            lambda progress: collect_stats(
                root,
                exts,
                args.include_root,
                ffprobe,
                progress=progress,
                include_tracks=args.per_track,
            ),
        )

    if not releases:
        if args.json:
            output = render_stats_json([], summary, root, include_tracks=args.per_track)
        elif args.csv:
            if args.per_track:
                output = render_stats_csv_tracks([], root)
            else:
                output = render_stats_csv([], root)
        else:
            if summary.scanned_audio_files > 0:
                output = "Audio files found, but metadata could not be read."
            else:
                output = "No audio files found."
        write_output_text(out_path, output)
        return 0

    groups = group_releases(releases, root)
    if args.json:
        output = render_stats_json(groups, summary, root, include_tracks=args.per_track)
        write_output_text(out_path, output)
        return 0
    if args.csv:
        if args.per_track:
            output = render_stats_csv_tracks(releases, root)
        else:
            output = render_stats_csv(releases, root)
        write_output_text(out_path, output)
        return 0

    output = render_stats_text(groups, summary, root, include_tracks=args.per_track)
    write_output_text(out_path, output)
    return 0



from __future__ import annotations

import logging
import shutil
from pathlib import Path

logger = logging.getLogger(__name__)


def ensure_root(root: Path) -> bool:
    if not root.exists() or not root.is_dir():
        logger.error("Error: '%s' is not a directory.", root)
        return False
    return True


def ensure_executable(name: str) -> bool:
    if shutil.which(name) is None:
        logger.error("Error: '%s' not found in PATH.", name)
        return False
    return True


def resolve_root(path_str: str) -> Path:
    return Path(path_str).expanduser().resolve()


def filter_existing_roots(roots: list[Path]) -> list[Path]:
    existing: list[Path] = []
    for root in roots:
        if root.exists() and root.is_dir():
            existing.append(root)
            continue
        logger.warning("Warning: skipping missing root: %s", root)
    return existing


def prepare_audio_root(path_str: str, *, skip_checks: bool) -> tuple[Path, int | None]:
    root = resolve_root(path_str)
    if skip_checks:
        return root, None
    if not ensure_root(root):
        return root, 2
    if not ensure_executable("ffprobe"):
        return root, 3
    return root, None


def is_within(path: Path, root: Path) -> bool:
    try:
        return path.resolve().is_relative_to(root.resolve())
    except AttributeError:
        resolved = path.resolve()
        root_resolved = root.resolve()
        return root_resolved == resolved or root_resolved in resolved.parents


def ensure_outside_roots(path: Path, roots: list[Path], label: str) -> bool:
    for root in roots:
        if is_within(path, root):
            logger.error("Error: %s cannot be inside the music root: %s", label, root)
            return False
    return True


def write_output_text(out_path: Path | None, text: str) -> None:
    if out_path is None:
        print(text)
        return
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(text, encoding="utf-8")


def write_output_lines(out_path: Path | None, lines: list[str]) -> None:
    text = "\n".join(lines)
    if out_path is None:
        for line in lines:
            print(line)
        return
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(text + ("\n" if text else ""), encoding="utf-8")



from __future__ import annotations

import json
import os
import shutil
from collections import Counter
from pathlib import Path
from typing import Any

from ..domain.dedupe import DedupeResult, TrackFingerprint, find_redundant_releases
from ..infra.fingerprint import iter_fingerprints
from ..infra.scan import iter_audio_files, release_root_for_path
from ..formatting.dedupe import dedupe_result_to_dict
from .dedupe_reporting import DedupeReportPaths, apply_actions, ensure_dir, print_summary, write_reports
from .progress import ProgressCallback


def _require_executable(cmd: str) -> bool:
    return shutil.which(cmd) is not None


def validate_actions(
    *, delete: bool, move_to: Path | None, require_action: bool
) -> tuple[int, str | None]:
    if delete and move_to is not None:
        return 2, "Error: --delete and --move-to cannot be used together"
    if require_action and not delete and move_to is None:
        return 2, "Error: --apply-plan requires --move-to or --delete"
    return 0, None


def write_plan(result: DedupeResult, plan_path: Path, roots: list[Path], exts: set[str]) -> None:
    data = dedupe_result_to_dict(result, roots=roots, exts=exts)
    plan_path.parent.mkdir(parents=True, exist_ok=True)
    plan_path.write_text(json.dumps(data, ensure_ascii=False, indent=2), encoding="utf-8")


def load_plan(plan_path: Path) -> dict[str, Any]:
    return json.loads(plan_path.read_text(encoding="utf-8"))


def apply_plan(
    plan_path: Path,
    *,
    move_to: Path | None,
    delete: bool,
    quiet: bool,
    progress: ProgressCallback | None = None,
) -> tuple[int, int, int]:
    code, err = validate_actions(delete=delete, move_to=move_to, require_action=True)
    if err:
        print(err)
        return code, 0, 0
    data = load_plan(plan_path)
    redundant = {Path(p) for p in data.get("redundant", [])}
    result = DedupeResult(
        redundant=redundant,
        duplicate_of={},
        contained_in={},
        unique_count={},
        sizes={},
        post_contained=[],
        unsafe=[],
    )
    moved, deleted_count = apply_actions(
        result,
        move_to=str(move_to) if move_to is not None else None,
        delete=delete,
        quiet=quiet,
        progress=progress,
    )
    return 0, moved, deleted_count


def run_dedupe(
    *,
    roots: list[Path],
    exts: set[str],
    out_dir: Path,
    jobs: int,
    move_to: Path | None,
    delete: bool,
    quiet: bool,
    progress: ProgressCallback | None = None,
    plan_out: Path | None = None,
) -> tuple[int, DedupeResult | None, DedupeReportPaths | None]:
    if not _require_executable("fpcalc"):
        print("Error: 'fpcalc' not found in PATH. Install chromaprint (fpcalc).", flush=True)
        return 2, None, None

    code, err = validate_actions(delete=delete, move_to=move_to, require_action=False)
    if err:
        print(err, flush=True)
        return code, None, None

    ensure_dir(out_dir)

    audio_files = list(iter_audio_files(roots, exts))
    if not audio_files:
        print("No audio files found.", flush=True)
        return 1, None, None

    if not quiet:
        print(f"Audio files found: {len(audio_files)}")
        print(f"Computing fpcalc fingerprints (jobs={jobs})...")

    if progress is not None:
        progress.start(len(audio_files))

    tsv_path = out_dir / "discog_audiofp.tsv"
    release_keys: dict[Path, Counter[TrackFingerprint]] = {}
    fingerprint_count = 0

    with tsv_path.open("w", encoding="utf-8") as f:
        for row in iter_fingerprints(audio_files, jobs, on_progress=progress.advance if progress else None):
            fingerprint_count += 1
            f.write(f"{row.duration}\t{row.fingerprint}\t{row.path.as_posix()}\n")
            rel = release_root_for_path(row.path, roots)
            if rel is None:
                continue
            release_keys.setdefault(rel, Counter())[TrackFingerprint(row.duration, row.fingerprint)] += 1

    if progress is not None:
        progress.finish()

    if fingerprint_count == 0:
        print("Error: fpcalc failed to process any files (check codecs/files).", flush=True)
        return 1, None, None

    result = find_redundant_releases(release_keys)
    paths = write_reports(result, out_dir)

    if plan_out is not None:
        write_plan(result, plan_out, roots, exts)

    if not quiet:
        print_summary(result, paths, out_dir)

    apply_actions(
        result,
        move_to=str(move_to) if move_to is not None else None,
        delete=delete,
        quiet=quiet,
        progress=progress,
    )

    return 0, result, paths


def default_jobs() -> int:
    return max(1, (os.cpu_count() or 2))



from __future__ import annotations

import argparse
import json
import tempfile
from pathlib import Path

from ..args import add_no_progress_arg, normalize_exts
from ..common import ensure_outside_roots, filter_existing_roots, is_within, write_output_lines, write_output_text
from ..progress import run_with_progress
from ...app.dedupe import apply_plan, default_jobs, load_plan, run_dedupe
from ...domain.constants import AUDIO_EXTS_DEFAULT
from ...formatting.dedupe import dedupe_result_to_dict, iter_dedupe_jsonl, render_dedupe_csv


def _protected_roots(roots: list[Path]) -> list[Path]:
    protected = list(roots)
    if len(roots) > 1:
        parents = {root.parent for root in roots}
        if len(parents) == 1:
            parent = parents.pop()
            if parent not in protected:
                protected.append(parent)
    return protected


def _is_safe_aux_path(path: Path, protected_roots: list[Path]) -> bool:
    return all(not is_within(path, root) for root in protected_roots)


def _default_out_dir(roots: list[Path], protected_roots: list[Path]) -> Path:
    cwd_candidate = (Path.cwd() / "_dedupe_reports").resolve()
    if _is_safe_aux_path(cwd_candidate, protected_roots):
        return cwd_candidate

    if len(roots) > 1:
        parents = {root.parent for root in roots}
        if len(parents) == 1:
            parent = parents.pop()
            sibling_candidate = (parent.parent / f"{parent.name}_dedupe_reports").resolve()
            if _is_safe_aux_path(sibling_candidate, protected_roots):
                return sibling_candidate

    root_label = roots[0].parent.name if len(roots) > 1 else roots[0].name
    return (Path(tempfile.gettempdir()) / f"{root_label}_dedupe_reports").resolve()


def add_parser(subparsers) -> argparse.ArgumentParser:
    """Register the dedupe subcommand parser."""
    parser = subparsers.add_parser(
        "dedupe",
        help="Find duplicate releases by audio fingerprint.",
        description="Find duplicate releases by audio content (Chromaprint/fpcalc).",
    )
    parser.add_argument(
        "--roots",
        nargs="*",
        default=["Albums", "Singles"],
        help="Root folders to scan (default: Albums Singles).",
    )
    parser.add_argument(
        "--ext",
        nargs="*",
        default=sorted(AUDIO_EXTS_DEFAULT),
        help="Audio extensions list (default: common formats).",
    )
    parser.add_argument(
        "--out-dir",
        default=None,
        help="Where to write reports (default: auto; uses ./_dedupe_reports when safe).",
    )
    parser.add_argument(
        "--jobs",
        type=int,
        default=default_jobs(),
        help="Parallelism for fpcalc (default: cpu_count).",
    )
    parser.add_argument(
        "--move-to",
        default=None,
        help="If set: move duplicate releases to the folder (no deletion).",
    )
    parser.add_argument(
        "--delete",
        action="store_true",
        help="If set: delete duplicate releases (dangerous).",
    )
    parser.add_argument(
        "--plan-out",
        default=None,
        help="Write a dedupe plan JSON (no deletion unless --move-to/--delete is set).",
    )
    parser.add_argument(
        "--apply-plan",
        default=None,
        help="Apply a previously generated plan JSON (skips scanning).",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="Output JSON report to stdout.",
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        help="Output CSV report to stdout.",
    )
    parser.add_argument(
        "--jsonl",
        action="store_true",
        help="Output JSON Lines report to stdout.",
    )
    parser.add_argument(
        "--output",
        default=None,
        help="Write JSON/CSV/JSONL output to a file instead of stdout.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Do not delete or move releases; only report.",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Reduce stdout output.",
    )
    add_no_progress_arg(parser)
    return parser


def run(args: argparse.Namespace) -> int:
    """Execute the dedupe command."""
    format_flags = [args.json, args.csv, args.jsonl]
    if sum(1 for flag in format_flags if flag) > 1:
        print("Error: --json, --csv, and --jsonl are mutually exclusive.")
        return 2
    if args.output and not any(format_flags):
        print("Error: --output requires --json, --csv, or --jsonl.")
        return 2
    if args.jsonl and args.apply_plan:
        print("Error: --jsonl cannot be used with --apply-plan.")
        return 2
    if args.dry_run and args.apply_plan:
        print("Error: --dry-run cannot be used with --apply-plan.")
        return 2

    if args.apply_plan and args.plan_out:
        print("Error: --apply-plan cannot be combined with --plan-out.")
        return 2

    if args.apply_plan:
        plan_path = Path(args.apply_plan).expanduser().resolve()
        plan_data = load_plan(plan_path)
        plan_roots = [Path(p).expanduser().resolve() for p in plan_data.get("roots", [])]
        protected_plan_roots = _protected_roots(plan_roots)
        move_to = Path(args.move_to).expanduser().resolve() if args.move_to else None
        if move_to is not None and protected_plan_roots:
            if not ensure_outside_roots(move_to, protected_plan_roots, "move target"):
                return 2
        out_path = Path(args.output).expanduser().resolve() if args.output else None
        if out_path is not None and protected_plan_roots:
            if not ensure_outside_roots(out_path, protected_plan_roots, "output file"):
                return 2
        quiet = args.quiet or args.json or args.csv or out_path is not None
        code, moved, deleted = run_with_progress(
            args.no_progress,
            quiet,
            "Applying dedupe plan",
            lambda progress: apply_plan(
                plan_path,
                move_to=move_to,
                delete=args.delete,
                quiet=quiet,
                progress=progress,
            ),
        )
        if code != 0:
            return code
        if args.json:
            output = json.dumps({"moved": moved, "deleted": deleted}, ensure_ascii=False)
        elif args.csv:
            output = f"moved,deleted\n{moved},{deleted}"
        else:
            return 0
        write_output_text(out_path, output)
        return 0

    roots = filter_existing_roots([Path(r).expanduser().resolve() for r in args.roots])
    if not roots:
        print("Error: no valid roots to scan.", flush=True)
        return 2
    protected_roots = _protected_roots(roots)
    exts = normalize_exts(args.ext, base_exts=set())
    out_dir = (
        Path(args.out_dir).expanduser().resolve()
        if args.out_dir is not None
        else _default_out_dir(roots, protected_roots)
    )
    move_to = Path(args.move_to).expanduser().resolve() if args.move_to else None
    plan_out = Path(args.plan_out).expanduser().resolve() if args.plan_out else None
    out_path = Path(args.output).expanduser().resolve() if args.output else None
    delete = args.delete

    if args.dry_run:
        move_to = None
        delete = False

    if not ensure_outside_roots(out_dir, protected_roots, "out dir"):
        return 2
    if plan_out is not None and not ensure_outside_roots(plan_out, protected_roots, "plan file"):
        return 2
    if out_path is not None and not ensure_outside_roots(out_path, protected_roots, "output file"):
        return 2
    if move_to is not None and not ensure_outside_roots(move_to, protected_roots, "move target"):
        return 2

    quiet = args.quiet or args.json or args.csv or args.jsonl or out_path is not None
    code, result, _ = run_with_progress(
        args.no_progress,
        quiet,
        "Fingerprinting audio",
        lambda progress: run_dedupe(
            roots=roots,
            exts=exts,
            out_dir=out_dir,
            jobs=args.jobs,
            move_to=move_to,
            delete=delete,
            quiet=quiet,
            progress=progress,
            plan_out=plan_out,
        ),
    )

    if code != 0:
        return code
    if result is None:
        return 1

    if args.json:
        output = json.dumps(dedupe_result_to_dict(result, roots=roots, exts=exts), ensure_ascii=False)
        write_output_text(out_path, output)
    elif args.csv:
        output = render_dedupe_csv(result)
        write_output_text(out_path, output)
    elif args.jsonl:
        write_output_lines(out_path, list(iter_dedupe_jsonl(result)))

    return 0



from __future__ import annotations

from pathlib import Path

from ..infra.scan import ReleaseScan, iter_release_scans


def list_release_scans(
    root: Path,
    exts: set[str],
    include_root: bool,
    *,
    sort: bool = True,
) -> list[ReleaseScan]:
    """Return release scans with optional sorting."""
    scans = list(iter_release_scans(root, exts, include_root))
    if sort:
        scans.sort(key=lambda item: item.path.as_posix().lower())
    return scans



from __future__ import annotations

import argparse
import logging
from pathlib import Path

from ..args import add_common_audio_args, add_no_progress_arg, normalize_exts
from ..common import ensure_outside_roots, prepare_audio_root, write_output_text
from ..progress import run_with_progress
from ...app.release import NoReadableAudioError, build_release_bbcode
from ...formatting.release import render_missing_assets_report

logger = logging.getLogger(__name__)


def add_parser(subparsers) -> argparse.ArgumentParser:
    """Register the release subcommand parser."""
    parser = subparsers.add_parser("release", help="Generate BBCode release template.")
    add_common_audio_args(parser, include_root=True)
    add_no_progress_arg(parser)
    parser.add_argument("--dr-dir", default=None, help="Directory with DR reports (e.g. *_dr.txt).")
    parser.add_argument(
        "--synthetic",
        action="store_true",
        help="Generate fake data (no ffprobe/files needed) to test output formatting.",
    )
    parser.add_argument("--no-cover", action="store_true", help="Disable cover upload to FastPic.")
    parser.add_argument("--lang", choices=["ru", "en"], default="ru", help="BBCode language (default: ru).")
    parser.add_argument(
        "--output",
        default=None,
        help="Write BBCode output to a file (default: ./<root>.txt).",
    )
    parser.add_argument(
        "--report-missing",
        nargs="?",
        const="missing_report.txt",
        default=None,
        help="Write a report of missing covers, skipped/failed cover uploads, and DR reports.",
    )
    return parser


def run(args: argparse.Namespace) -> int:
    """Execute the release command."""
    root, err = prepare_audio_root(args.root, skip_checks=args.synthetic)
    if err is not None:
        return err

    exts = normalize_exts(args.ext)

    dr_dir: Path | None = None
    if args.dr_dir is not None:
        dr_dir = Path(args.dr_dir).expanduser().resolve()
        if not dr_dir.exists() or not dr_dir.is_dir():
            logger.warning("Warning: --dr-dir path is not a directory: %s", dr_dir)
            dr_dir = None

    try:
        if args.synthetic:
            result = build_release_bbcode(
                root,
                exts,
                args.include_root,
                dr_dir=dr_dir,
                test_mode=True,
                no_cover=args.no_cover,
                lang=args.lang,
            )
        else:
            result = run_with_progress(
                args.no_progress,
                False,
                "Reading audio metadata",
                lambda progress: build_release_bbcode(
                    root,
                    exts,
                    args.include_root,
                    dr_dir=dr_dir,
                    test_mode=False,
                    no_cover=args.no_cover,
                    lang=args.lang,
                    progress=progress,
                ),
            )
    except NoReadableAudioError as exc:
        print(str(exc))
        return 0

    if result is None:
        print("No audio files found.")
        return 0

    if args.output is not None:
        out_path = Path(args.output).expanduser().resolve()
    else:
        out_path = Path.cwd() / f"{root.name}.txt"
    if not ensure_outside_roots(out_path, [root], "output file"):
        return 2
    write_output_text(out_path, result.bbcode)
    print(f"Wrote release template: {out_path}")

    if args.report_missing is not None:
        report_path = Path(args.report_missing).expanduser().resolve()
        if not ensure_outside_roots(report_path, [root], "missing report file"):
            return 2
        report_text = render_missing_assets_report(result, root, dr_dir=dr_dir)
        write_output_text(report_path, report_text)
        print(f"Wrote missing report: {report_path}")
    return 0



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from ..domain.normalize import (
    NormalizationInput,
    NormalizationPlan,
    NormalizationSkip,
    build_normalization_plan,
)
from ..domain.tags import append_release_metadata_values, most_common_str, select_release_metadata
from ..domain.utils import parse_year_from_text
from ..infra.ffprobe import FfprobeClient, TagsReader
from ..infra.scan import ReleaseScan
from .progress import ProgressCallback
from .scan import list_release_scans

TAG_SAMPLE_LIMIT = 8
TAG_MIN_SAMPLES = 3
TAG_MAJORITY = 0.6


@dataclass(frozen=True)
class NormalizationInputs:
    inputs: list[NormalizationInput]
    single_mode: bool


def _is_stable(values: list[str]) -> bool:
    if len(values) < TAG_MIN_SAMPLES:
        return False
    top = most_common_str(values)
    if not top:
        return False
    count = sum(1 for v in values if v == top)
    return (count / len(values)) >= TAG_MAJORITY


def _sample_tags(
    audio_files: list[Path],
    ffprobe: TagsReader,
    progress: ProgressCallback | None,
) -> list[dict[str, str]]:
    tags_list: list[dict[str, str]] = []
    album_values: list[str] = []
    album_artist_values: list[str] = []
    artist_values: list[str] = []

    for index, path in enumerate(audio_files):
        if index >= TAG_SAMPLE_LIMIT:
            break
        tags = ffprobe.get_tags(path)
        tags_list.append(tags)
        if progress is not None:
            progress.advance()

        append_release_metadata_values(tags, album_values, album_artist_values, artist_values)

        album_ok = _is_stable(album_values)
        artist_source = album_artist_values if album_artist_values else artist_values
        artist_ok = _is_stable(artist_source)

        if album_ok and artist_ok:
            break

    return tags_list


def _resolve_release_scans(root: Path, exts: set[str]) -> tuple[list[ReleaseScan], bool]:
    release_data = list_release_scans(root, exts, include_root=True, sort=True)
    if not release_data:
        return [], True
    single_mode = len(release_data) == 1
    if not single_mode:
        release_data = [item for item in release_data if item.path != root]
    return release_data, single_mode


def collect_normalization_inputs(
    root: Path,
    exts: set[str],
    ffprobe: TagsReader,
    progress: ProgressCallback | None,
) -> NormalizationInputs:
    release_data, single_mode = _resolve_release_scans(root, exts)
    if not release_data:
        return NormalizationInputs(inputs=[], single_mode=single_mode)

    total_files = sum(min(len(item.audio_files), TAG_SAMPLE_LIMIT) for item in release_data)
    if progress is not None:
        progress.start(total_files)

    inputs: list[NormalizationInput] = []
    for item in release_data:
        tags_list = _sample_tags(item.audio_files, ffprobe, progress)
        artist, album = select_release_metadata(tags_list)
        year = parse_year_from_text(item.path.name)
        inputs.append(NormalizationInput(path=item.path, artist=artist, album=album, year=year))

    return NormalizationInputs(inputs=inputs, single_mode=single_mode)


def plan_normalization(
    root: Path,
    exts: set[str],
    tag_reader: TagsReader | None = None,
    progress: ProgressCallback | None = None,
) -> NormalizationPlan:
    """Build a normalization plan and skip unsafe targets."""
    ffprobe = tag_reader or FfprobeClient()
    inputs = collect_normalization_inputs(root, exts, ffprobe, progress)
    if not inputs.inputs:
        if progress is not None:
            progress.finish()
        return NormalizationPlan(actions=[], skipped=[])

    plan = build_normalization_plan(inputs.inputs, single_mode=inputs.single_mode)
    if not plan.actions:
        if progress is not None:
            progress.finish()
        return plan

    actions = []
    skipped = list(plan.skipped)
    for action in plan.actions:
        if action.target.exists() and action.target != action.source:
            skipped.append(
                NormalizationSkip(path=action.source, reason=f"target exists: {action.target}")
            )
            continue
        actions.append(action)

    if progress is not None:
        progress.finish()

    return NormalizationPlan(actions=actions, skipped=skipped)


def apply_normalization(
    plan: NormalizationPlan,
    progress: ProgressCallback | None = None,
) -> int:
    """Apply rename actions from a normalization plan."""
    total_actions = len(plan.actions)
    if progress is not None:
        progress.set_description("Renaming releases")
        progress.start(total_actions)

    for index, action in enumerate(plan.actions, start=1):
        if progress is not None:
            progress.set_description(
                f"Renaming releases ({index}/{total_actions}): {action.source.name} -> {action.target.name}"
            )
        action.source.rename(action.target)
        if progress is not None:
            progress.advance()
    return len(plan.actions)



from __future__ import annotations

import csv
from io import StringIO
from typing import Iterable


def render_csv_rows(rows: Iterable[Iterable[object]]) -> str:
    buffer = StringIO(newline="")
    writer = csv.writer(buffer, lineterminator="\n")
    writer.writerows(rows)
    return buffer.getvalue().rstrip("\n")



from __future__ import annotations

import re
from pathlib import Path

_TRACK_NUM_RE = re.compile(r"^\s*(\d{1,3})\s*([.\-_\s]+)\s*(.*)$")


def _parse_track_num(path: Path) -> int | None:
    match = _TRACK_NUM_RE.match(path.stem)
    if not match:
        return None
    return int(match.group(1))


def _track_sort_key(path: Path) -> tuple[int, int, str]:
    track_num = _parse_track_num(path)
    if track_num is not None:
        return (0, track_num, path.name.lower())
    return (1, 10**9, path.name.lower())


def build_tracklist_lines(audio_files: list[Path], *, sort: bool = True) -> list[str]:
    """
    Builds a tracklist from file names.

    Logic is unchanged:
    - If the name starts with a track number, keep it.
    - Otherwise, auto-number starting at 01.
    """
    lines: list[str] = []
    auto_n = 1
    used_numbers: set[int] = set()

    files = sorted(audio_files, key=_track_sort_key) if sort else audio_files
    for f in files:
        stem = f.stem
        m = _TRACK_NUM_RE.match(stem)
        if m:
            num = m.group(1)
            num_int = int(num)
            title = m.group(3).strip() or stem
            width = 3 if len(num) >= 3 else 2
            num_fmt = f"{num_int:0{width}d}"
            used_numbers.add(num_int)
            auto_n = max(auto_n, num_int + 1)
        else:
            while auto_n in used_numbers:
                auto_n += 1
            num_fmt = f"{auto_n:02d}"
            title = stem.strip() or stem
            used_numbers.add(auto_n)
            auto_n += 1

        title = re.sub(r"\s+", " ", title).strip()
        lines.append(f"{num_fmt}. {title}")

    return lines



from __future__ import annotations

from typing import Protocol


class ProgressCallback(Protocol):
    def start(self, total: int) -> None: ...

    def advance(self, step: int = 1) -> None: ...

    def set_description(self, description: str) -> None: ...

    def finish(self) -> None: ...


class NullProgress:
    def start(self, total: int) -> None:
        return None

    def advance(self, step: int = 1) -> None:
        return None

    def set_description(self, description: str) -> None:
        return None

    def finish(self) -> None:
        return None



from __future__ import annotations

from dataclasses import dataclass, field
import logging
from pathlib import Path

from ..domain.grouping import ReleaseBBCodeItem, group_bbcode_releases
from ..domain.models import ReleaseBBCode
from ..domain.utils import (
    bit_label,
    codec_label,
    format_hhmmss,
    group_key,
    parse_release_title_and_year,
    sr_label,
)
from ..formatting.bbcode import make_release_bbcode, make_single_release_bbcode
from ..formatting.tracklist import build_tracklist_lines
from ..infra.cover import FastPicCoverUploader, find_cover_jpg, requests as cover_requests
from ..infra.dr import build_dr_index, find_dr_text_for_release
from ..infra.ffprobe import FfprobeClient
from .progress import ProgressCallback
from .stats import collect_stats, collect_synthetic_stats

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ReleaseBuildResult:
    bbcode: str
    total_releases: int
    missing_covers: list[Path]
    missing_drs: list[Path]
    dr_checked: bool
    failed_cover_uploads: list[Path] = field(default_factory=list)
    cover_uploads_skipped: bool = False


class NoReadableAudioError(RuntimeError):
    pass


def _progress_label(stage: str, index: int, total: int, release_name: str) -> str:
    return f"{stage} ({index}/{total}): {release_name}"


def build_release_bbcode(
    root: Path,
    exts: set[str],
    include_root: bool,
    *,
    dr_dir: Path | None,
    test_mode: bool,
    no_cover: bool,
    lang: str,
    progress: ProgressCallback | None = None,
) -> ReleaseBuildResult | None:
    if test_mode:
        releases, summary = collect_synthetic_stats(root)
    else:
        ffprobe = FfprobeClient()
        releases, summary = collect_stats(root, exts, include_root, ffprobe, progress=progress)

    if not releases:
        if summary.scanned_audio_files > 0:
            raise NoReadableAudioError("Audio files found, but metadata could not be read.")
        return None

    dr_index: dict[str, Path] = {}
    if dr_dir is not None:
        dr_index = build_dr_index(dr_dir)
        dr_checked = True
    else:
        dr_checked = False

    cover_uploader: FastPicCoverUploader | None = None
    cover_uploads_skipped = False
    if not test_mode and not no_cover:
        if cover_requests is None:
            logger.warning("Warning: 'requests' not installed; skipping FastPic cover uploads.")
            cover_uploads_skipped = True
        else:
            cover_uploader = FastPicCoverUploader(resize_to=500)
    elif not test_mode and no_cover:
        cover_uploads_skipped = True

    year_range = None
    if summary.all_years:
        y_min, y_max = min(summary.all_years), max(summary.all_years)
        year_range = f"{y_min}-{y_max}" if y_min != y_max else f"{y_min}"

    items: list[ReleaseBBCodeItem] = []
    missing_covers: list[Path] = []
    missing_drs: list[Path] = []
    failed_cover_uploads: list[Path] = []
    total_releases = len(releases)
    if progress is not None:
        progress.set_description("Preparing releases")
        progress.start(total_releases)

    for index, rel in enumerate(releases, start=1):
        if progress is not None:
            progress.set_description(_progress_label("Preparing releases", index, total_releases, rel.path.name))
        rel_path = rel.path.relative_to(root)
        group = group_key(rel_path)

        folder_name = rel.path.name
        title, year = parse_release_title_and_year(folder_name)
        tracklist = build_tracklist_lines(rel.audio_files, sort=True)

        dr_text = None
        if test_mode:
            dr_text = rel.dr_text
        elif dr_dir is not None:
            dr_text = find_dr_text_for_release(folder_name, dr_dir, dr_index)
        if dr_checked and dr_text is None:
            missing_drs.append(rel.path)

        cover_url = None
        cover_path = find_cover_jpg(rel.path)
        if cover_path is None:
            missing_covers.append(rel.path)
        elif cover_uploader is not None:
            try:
                if progress is not None:
                    progress.set_description(_progress_label("Uploading covers", index, total_releases, rel.path.name))
                cover_url = cover_uploader.upload(cover_path)
            except Exception as exc:
                failed_cover_uploads.append(rel.path)
                logger.warning("Warning: cover upload failed for %s: %s", cover_path, exc)

        items.append(
            ReleaseBBCodeItem(
                group=group,
                release=ReleaseBBCode(
                    title=title,
                    year=year,
                    duration=format_hhmmss(rel.duration_seconds),
                    tracklist=tracklist,
                    dr=dr_text,
                    cover_url=cover_url,
                ),
            )
        )
        if progress is not None:
            progress.advance()

    groups = group_bbcode_releases(items)

    if total_releases == 1:
        single_release = groups[0].releases[0] if groups and groups[0].releases else None
        if single_release is None:
            logger.warning("Warning: no releases found for BBCode generation.")
            return None

        bbcode = make_single_release_bbcode(
            root_name=root.name,
            year_range=year_range,
            overall_codec=codec_label(summary.total_exts),
            release=single_release,
            lang=lang,
        )
    else:
        bbcode = make_release_bbcode(
            root_name=root.name,
            year_range=year_range,
            total_duration=format_hhmmss(summary.total_seconds),
            overall_codec=codec_label(summary.total_exts),
            overall_bit=bit_label(summary.total_bit),
            overall_sr=sr_label(summary.total_sr),
            grouped_releases=groups,
            lang=lang,
        )

    return ReleaseBuildResult(
        bbcode=bbcode,
        total_releases=total_releases,
        missing_covers=missing_covers,
        missing_drs=missing_drs,
        dr_checked=dr_checked,
        failed_cover_uploads=failed_cover_uploads,
        cover_uploads_skipped=cover_uploads_skipped,
    )
